
package visitor;

import syntaxtree.And;
import syntaxtree.ArrayAssign;
import syntaxtree.ArrayLength;
import syntaxtree.ArrayLookup;
import syntaxtree.Assign;
import syntaxtree.Block;
import syntaxtree.BooleanType;
import syntaxtree.Call;
import syntaxtree.ClassDeclExtends;
import syntaxtree.ClassDeclSimple;
import syntaxtree.False;
import syntaxtree.Formal;
import syntaxtree.Identifier;
import syntaxtree.IdentifierExp;
import syntaxtree.IdentifierType;
import syntaxtree.If;
import syntaxtree.IntArrayType;
import syntaxtree.IntegerLiteral;
import syntaxtree.IntegerType;
import syntaxtree.LessThan;
import syntaxtree.MainClass;
import syntaxtree.MethodDecl;
import syntaxtree.MethodDeclList;
import syntaxtree.Minus;
import syntaxtree.NewArray;
import syntaxtree.NewObject;
import syntaxtree.Not;
import syntaxtree.Plus;
import syntaxtree.Print;
import syntaxtree.Program;
import syntaxtree.StatementList;
import syntaxtree.This;
import syntaxtree.Times;
import syntaxtree.True;
import syntaxtree.VarDecl;
import syntaxtree.VarDeclList;
import syntaxtree.While;

public class PrettyPrintVisitor implements GenVisitor<Integer, Void> {

   public static Integer DEFAULT_INITIAL_INDENTATION = 0;
   public static Integer TAB_WIDTH = 2;

   private void indent(int columns) {

      StringBuilder s = new StringBuilder(columns);
      for (int i = 0; i < columns; ++i) {
         s.append(" ");
      }

      System.out.print(s.toString());
   }

   private void print(int columns, String str) {
      indent(columns);
      System.out.print(str);
   }

   private void println(int columns, String str) {
      print(columns, str);
      System.out.println();
   }

   private void nl() {
      System.out.println();
   }

   // MainClass m;
   // ClassDeclList cl;
   public Void visit(Program n, Integer indentation) {
      n.m.accept(this, indentation);
      for (int i = 0; i < n.cl.size(); i++) {
         System.out.println();
         n.cl.elementAt(i).accept(this, indentation);
      }

      return null;
   }

   // Identifier i1,i2;
   // Statement s;
   public Void visit(MainClass n, Integer indentation) {

      indent(indentation);
      System.out.print("class ");
      n.i1.accept(this, indentation);
      System.out.println(" {");

      indent(indentation + TAB_WIDTH);
      System.out.print("public static void main (String [] ");
      n.i2.accept(this, indentation);
      System.out.println(") {");

      n.s.accept(this, indentation + TAB_WIDTH + TAB_WIDTH);

      indent(indentation + TAB_WIDTH);
      System.out.println("}");

      indent(indentation);
      System.out.println("}");

      return null;
   }

   // Identifier i;
   // VarDeclList vl;
   // MethodDeclList ml;
   public Void visit(ClassDeclSimple n, Integer indentation) {
      indent(indentation);
      System.out.print("class ");
      n.i.accept(this, indentation);
      System.out.println(" { ");
      printVarDecls(n.vl, indentation);
      nl();

      printMethodDeclList(n.ml, indentation);
      System.out.println();
      System.out.println("}");
      nl();

      return null;
   }

   // Identifier i;
   // Identifier j;
   // VarDeclList vl;
   // MethodDeclList ml;
   public Void visit(ClassDeclExtends n, Integer indentation) {
      System.out.print("class ");
      n.i.accept(this, indentation);
      System.out.println(" extends ");
      n.j.accept(this, indentation);
      System.out.println(" { ");

      printVarDecls(n.vl, indentation);
      nl();

      printMethodDeclList(n.ml, indentation );

      System.out.println();
      System.out.println("}");

      return null;
   }

   // Type t;
   // Identifier i;
   public Void visit(VarDecl n, Integer indentation) {
      indent(indentation);
      n.t.accept(this, indentation);
      System.out.print(" ");
      n.i.accept(this, indentation);
      System.out.print(";");
      nl();

      return null;
   }

   // Type t;
   // Identifier i;
   // FormalList fl;
   // VarDeclList vl;
   // StatementList sl;
   // Exp e;
   public Void visit(MethodDecl n, Integer indentation) {
      indent(indentation);
      System.out.print("public ");
      n.t.accept(this, indentation);
      System.out.print(" ");
      n.i.accept(this, indentation);
      System.out.print(" (");
      for (int i = 0; i < n.fl.size(); i++) {
         n.fl.elementAt(i).accept(this, indentation);
         if (i + 1 < n.fl.size()) {
            System.out.print(", ");
         }
      }
      System.out.println(") { ");

      printVarDecls(n.vl, indentation);
      nl();
      printStatementList(n.sl, indentation);

      nl();
      indent(indentation + TAB_WIDTH);
      System.out.print("return ");
      n.e.accept(this, null);
      System.out.println(";");
      indent(indentation);
      System.out.print("}");

      nl();
      return null;
   }

   private void printVarDecls(VarDeclList varDecls, Integer indentation) {
      for (int i = 0; i < varDecls.size(); i++) {
         varDecls.elementAt(i).accept(this, indentation + TAB_WIDTH);
      }
   }

   private void printStatementList(StatementList statements, Integer indentation) {
      for (int i = 0; i < statements.size(); i++) {
         statements.elementAt(i).accept(this, indentation + TAB_WIDTH);
      }
   }

   private void printMethodDeclList(MethodDeclList methodDecls,
                                    Integer indentation) {
      for (int i = 0; i < methodDecls.size(); i++) {
         methodDecls.elementAt(i).accept(this, indentation + TAB_WIDTH);
         nl();
      }
   }

   // Type t;
   // Identifier i;
   public Void visit(Formal n, Integer indentation) {
      n.t.accept(this, indentation);
      System.out.print(" ");
      n.i.accept(this, indentation);

      return null;
   }

   public Void visit(IntArrayType n, Integer indentation) {
      System.out.print("int []");
      return null;
   }

   public Void visit(BooleanType n, Integer indentation) {
      System.out.print("boolean");
      return null;
   }

   public Void visit(IntegerType n, Integer indentation) {
      System.out.print("int");
      return null;
   }

   // String s;
   public Void visit(IdentifierType n, Integer indentation) {
      System.out.print(n.s);
      return null;
   }

   // StatementList sl;
   public Void visit(Block n, Integer indentation) {
      print(indentation, "{ ");
      nl();
      printStatementList(n.sl, indentation + TAB_WIDTH);
      print(indentation, "}");
      nl();
      return null;
   }

   // Exp e;
   // Statement s1,s2;
   public Void visit(If n, Integer indentation) {

      print(indentation, "if (");
      n.e.accept(this, indentation);
      System.out.print(") ");
      nl();

      // Print Statement - they end with ;\n
      n.s1.accept(this, indentation + TAB_WIDTH);

      print(indentation, "else");
      nl();
      // Print Statement - they end with ;\n
      n.s2.accept(this, indentation + TAB_WIDTH);

      return null;
   }

   // Exp e;
   // Statement s;
   public Void visit(While n, Integer indentation) {
      indent(indentation);
      System.out.print("while (");
      n.e.accept(this, indentation + TAB_WIDTH);
      System.out.print(") ");
      nl();
      n.s.accept(this, indentation + TAB_WIDTH);
      return null;
   }

   // Exp e;
   public Void visit(Print n, Integer indentation) {
      indent(indentation);
      System.out.print("System.out.println(");
      n.e.accept(this, indentation);
      System.out.print(");");
      nl();
      return null;
   }

   // Identifier i;
   // Exp e;
   public Void visit(Assign n, Integer indentation) {
      indent(indentation);
      n.i.accept(this, indentation);
      System.out.print(" = ");
      n.e.accept(this, indentation);
      System.out.print(";");
      nl();
      return null;
   }

   // Identifier i;
   // Exp e1,e2;
   public Void visit(ArrayAssign n, Integer indentation) {
      indent(indentation);
      n.i.accept(this, indentation);
      System.out.print("[");
      n.e1.accept(this, indentation);
      System.out.print("] = ");
      n.e2.accept(this, indentation);
      System.out.print(";");
      nl();
      return null;
   }

   // Exp e1,e2;
   public Void visit(And n, Integer indentation) {
      System.out.print("(");
      n.e1.accept(this, indentation);
      System.out.print(" && ");
      n.e2.accept(this, indentation);
      System.out.print(")");
      return null;
   }

   // Exp e1,e2;
   public Void visit(LessThan n, Integer indentation) {
      System.out.print("(");
      n.e1.accept(this, indentation);
      System.out.print(" < ");
      n.e2.accept(this, indentation);
      System.out.print(")");
      return null;
   }

   // Exp e1,e2;
   public Void visit(Plus n, Integer indentation) {
      System.out.print("(");
      n.e1.accept(this, indentation);
      System.out.print(" + ");
      n.e2.accept(this, indentation);
      System.out.print(")");
      return null;
   }

   // Exp e1,e2;
   public Void visit(Minus n, Integer indentation) {
      System.out.print("(");
      n.e1.accept(this, indentation);
      System.out.print(" - ");
      n.e2.accept(this, indentation);
      System.out.print(")");
      return null;
   }

   // Exp e1,e2;
   public Void visit(Times n, Integer indentation) {
      System.out.print("(");
      n.e1.accept(this, indentation);
      System.out.print(" * ");
      n.e2.accept(this, indentation);
      System.out.print(")");
      return null;
   }

   // Exp e1,e2;
   public Void visit(ArrayLookup n, Integer indentation) {
      n.e1.accept(this, indentation);
      System.out.print("[");
      n.e2.accept(this, indentation);
      System.out.print("]");
      return null;
   }

   // Exp e;
   public Void visit(ArrayLength n, Integer indentation) {
      n.e.accept(this, indentation);
      System.out.print(".length");
      return null;
   }

   // Exp e;
   // Identifier i;
   // ExpList el;
   // TODO
   public Void visit(Call n, Integer indentation) {
      n.e.accept(this, indentation);
      System.out.print(".");
      n.i.accept(this, indentation);
      System.out.print("(");
      for (int i = 0; i < n.el.size(); i++) {
         n.el.elementAt(i).accept(this, indentation);
         if (i + 1 < n.el.size()) {
            System.out.print(", ");
         }
      }
      System.out.print(")");
      return null;
   }

   // int i;
   public Void visit(IntegerLiteral n, Integer indentation) {
      System.out.print(n.i);
      return null;
   }

   public Void visit(True n, Integer indentation) {
      System.out.print("true");
      return null;
   }

   public Void visit(False n, Integer indentation) {
      System.out.print("false");
      return null;
   }

   // String s;
   public Void visit(IdentifierExp n, Integer indentation) {
      System.out.print(n.s );
      return null;
   }

   public Void visit(This n, Integer indentation) {
      System.out.print("this");
      return null;
   }

   // Exp e;
   public Void visit(NewArray n, Integer indentation) {
      System.out.print("new int [");
      n.e.accept(this, indentation);
      System.out.print("]");
      return null;
   }

   // Identifier i;
   public Void visit(NewObject n, Integer indentation) {
      System.out.print("new ");
      System.out.print(n.i.s);
      System.out.print("()");
      return null;
   }

   // Exp e;
   public Void visit(Not n, Integer indentation) {
      System.out.print("!");
      n.e.accept(this, indentation);
      return null;
   }

   // String s;
   public Void visit(Identifier n, Integer indentation) {
      System.out.print(n.s+ "{" + n.beginLine + "," + n.beginColumn + "}");
      return null;
   }
}
